*************************************************************************
*	This file contains 2 FORTRAN 77 routines for converting between
*	Julian and calendar dates. Routine JTODMY converst from 
*	Julian day number to calendar date (Gregorian); DMYTOJ
*	goes the other way round. Algorithms obtained from 
*	http://hermetic.magnet.ch/cal_stud/jdn.htm on 28th November
*	2001 (Julian day 2452242).
*************************************************************************
      SUBROUTINE JTODMY(JULIAN,D0,M0,Y0,DD,MM,YY,IFAIL)
*
*     Converts a date in JULIAN format, where day 0 is D0/M0/Y0, 
*     to DD/MM/YY in the Gregorian calendar. If Y0, M0 and D0 are all 
*     zero, the origin defaults to 24th November, 4713BC (which is when 
*     the calendar starts!). The origin is actually 1st January 4712BC 
*     according to the *Julian* calendar. This algorithm is only valid
*     for dates after 1st March, 4900BC. Algorithm due to Henry F. Fliegel
*     and Thomas C. Van Flandern. Tested against Splus version, and agrees.
*
      INTEGER JULIAN,Y0,M0,D0,YY,MM,DD,IFAIL
      INTEGER L,N,I,J,JD

      IFAIL = 0

      IF ((Y0.EQ.0).AND.(M0.EQ.0).AND.(D0.EQ.0)) THEN
       JD = JULIAN
      ELSE
       CALL DMYTOJ(D0,M0,Y0,JD,IFAIL)
       IF (IFAIL.NE.0) RETURN
       JD = JULIAN + JD
      ENDIF
      L = JD + 68569
      N = (4*L) / 146097
      L = L - ((146097*N) + 3) / 4
      I = (4000 * (L+1)) / 1461001
      L = L - ((1461*I) / 4) + 31
      J = (80*L)/2447
      DD = L - ((2447*J) / 80)
      L = J/11
      MM = J + 2 - (12*L)
      YY = (100 * (N-49)) + I + L

      IF (((10000*YY)+(100*MM)+DD).LT.-49000301) THEN
       IFAIL = 1
       WRITE(*,1)
       RETURN
      ENDIF

 1    FORMAT('****ERROR**** in Julian-to-date conversion: ',/,
     +     'Algorithm only works for dates before 1st March 4900BC!')
      END
*************************************************************************
*************************************************************************
*************************************************************************
      SUBROUTINE DMYTOJ(DD,MM,YY,JULIAN,IFAIL)
*
*     Converts date DD/MM/YY (on the Gregorian calendar) into a 
*     Julian day number (origin: 24th November 4713BC). This algorithm
*     is valid only for dates after 1st March, 4800BC. Algorithm due 
*     to Henry F. Fliegel and Thomas C. Van Flandern. Tested against 
*     Splus version, and agrees.
*
      INTEGER YY,MM,DD,JULIAN,IFAIL

      IFAIL = 0
      IF (YY.EQ.0) THEN
       IFAIL = 1
       WRITE(*,1)
       RETURN
      ELSEIF (((10000*YY)+(100*MM)+DD).LT.-48000301) THEN
       IFAIL = 2
       WRITE(*,2)
       RETURN
      ENDIF

      JULIAN = (1461 * (YY + 4800 + (MM - 14) / 12)) / 4 +
     +         (367 * (MM - 2 - 12 * ((MM - 14 ) / 12))) / 12 -
     +         (3 * ((YY + 4900 + (MM - 14) / 12) / 100)) / 4 +
     +         DD - 32075      

 1    FORMAT('****ERROR**** in date-to-Julian conversion: ',
     +     'Gregorian year 0 doesn''t exist!')
 2    FORMAT('****ERROR**** in date-to-Julian conversion: ',/,
     +     'Algorithm only works for dates before 1st March 4800BC!')
      END


