**************************************************************************
*        UCL 2002-2010
**************************************************************************
      BLOCK DATA DSCTXT
**************************************************************************
*       Initializes descriptive text relating to predictors etc. Whenever 
*       a new function or predictor is added, the corresponding entry 
*       should be updated in this block so that program output corresponds 
*       to what the thing is actually doing.
**************************************************************************
      COMMON /DESCRIBE/ SXPTXT,MOTXT,TRNDTXT,TRPTXT,DYTXT,
     +			SPATXT,SPPTXT,CORTXT,PWTTXT,GLBTXT
      INTEGER MXP,I,J

      PARAMETER (MXP=80)

      CHARACTER*70 MOTXT(MXP),TRNDTXT(MXP)
      CHARACTER*70 DYTXT(MXP),PWTTXT(MXP,3)
      CHARACTER*70 TRPTXT(MXP,3),SXPTXT(MXP,3)
      CHARACTER*70 SPATXT(0:MXP),SPPTXT(MXP,4),CORTXT(MXP)
      CHARACTER*70 GLBTXT(MXP)
 
      DATA ((SXPTXT(I,J),J=1,3),I=1,3) /
     +'Parameter of transformation','NOT USED','NOT USED',
     +'Value of a','NOT USED','NOT USED',
     +'Location parameter','Scale parameter','NOT USED'/
      DATA ((SXPTXT(I,J),I=11,30),J=1,3) /
     +20*'Lower limit for Fourier representation',
     +20*'Upper limit for Fourier representation',20*'NOT USED'/
      DATA ((SXPTXT(I,J),I=31,40),J=1,3) /
     +10*'Lower limit for polynomial representation',
     +10*'Upper limit for polynomial representation',10*'NOT USED'/

      DATA ((PWTTXT(I,J),J=1,3),I=2,3) /
     +'Exponential decay rate','NOT USED','NOT USED',
     +'Offset for site attribute 1','Offset for site attribute 2',
     +'Exponential decay rate'/

      DATA (MOTXT(I),I=1,4) /
     +'Monthly seasonal effect, cosine component',
     +'Monthly seasonal effect, sine component',
     +'Monthly half-year cycle, cosine component',
     +'Monthly half-year cycle, sine component'/
      DATA (MOTXT(I),I=11,22) /
     +'January indicator','February indicator','March indicator',
     +'April indicator','May indicator','June indicator',
     +'July indicator','August indicator','September indicator',
     +'October indicator','November indicator','December indicator' /

      DATA (TRNDTXT(I),I=1,3) /
     +'Linear (0.1 per year, 0 in 1950)',
     +'Linear (0.1 per year) after year Y',
     +'Cyclical'/

      DATA ((TRPTXT(I,J),J=1,3),I=1,3) /
     +'NOT USED','NOT USED','NOT USED',
     +'Date of changepoint (year)','NOT USED','NOT USED',
     +'Cycle length (years)','Date of cycle minimum','NOT USED'/

      DATA (DYTXT(I),I=1,10) /
     +'Y[t-1]',
     +'Y[t-2]',
     +'Y[t-3]',
     +'Y[t-4]',
     +'Y[t-5]',
     +'Y[t-6]',
     +'Y[t-7]',
     +'Y[t-8]',
     +'Y[t-9]',
     +'Y[t-10]'/
      DATA (DYTXT(I),I=21,24) /
     +'Daily seasonal effect, cosine component',
     +'Daily seasonal effect, sine component',
     +'Daily half-year cycle, cosine component',
     +'Daily half-year cycle, sine component'/
      DATA (DYTXT(I),I=31,42) /
     +'Smooth January effect','Smooth February effect',
     +'Smooth March effect','Smooth April effect',
     +'Smooth May effect','Smooth June effect',
     +'Smooth July effect','Smooth August effect',
     +'Smooth September effect','Smooth October effect',
     +'Smooth November effect','Smooth December effect' /
*
*	For spatial structure, default equates to independence.
*	Positions 1-20 are used for amounts model (correlation 
*	structure of Anscombe residuals), 21-40 are for wet/dry 
*	model.
*
      DATA (SPATXT(I),I=0,6) /'Independence',
     +'Empirical correlations for ',
     +'Constant inter-site correlations for',
     +'Exponential correlation function: exp[-phi*d] for',
     +'Thresholded exp. corr. fn: a + (1-a)exp[-phi*d] for',
     +'Powered exp. corr. fn: exp[-phi*(d^k)] for',
     +'Thresholded pow. exp. corr. fn: a + (1-a)exp[-phi*(d^k)] for'/
      DATA SPPTXT(2,1) /'Inter-site correlation'/
      DATA SPPTXT(3,1) /'Correlation decay rate, phi'/
      DATA (SPPTXT(4,I),I=1,2) /'Correlation decay rate, phi',
     +              'Limiting correlation at large distances, a'/
      DATA (SPPTXT(5,I),I=1,2) /'Correlation decay rate, phi',
     +                          'Power of distance, k'/
      DATA (SPPTXT(6,I),I=1,3) /'Correlation decay rate, phi',
     +              'Power of distance, d',
     +              'Limiting correlation at large distances, a'/
      DATA (SPATXT(I),I=21,22) /
     +'Conditional independence given ''wet/dry'' weather state',
     +'Beta-Binomial distribution for number of wet sites'/
      DATA SPPTXT(21,1) /'Increase in logit on a ''wet'' day'/
      DATA (SPPTXT(22,I),I=1,3) /
     +'Shape parameter of Beta-Binomial distribution',
     +'Emergency shrinkage parameter',
     +'Print errors? (0=no, else yes)'/
*
*	Text describing the quantities to which correlation 
*       structures correspond
*
      DATA CORTXT(1) /'latent Gaussian variables'/
      DATA CORTXT(11) /'Anscombe residuals'/
*
*     Text for global quantities
*
      DATA (GLBTXT(I),I=1,1) /'Threshold for small positive values'/

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE ATTRXFM(TITLE,SITINF,NSITES,ATTRTXT,NATTR,
     +				CODES,XFM,FOUIDX,LEGIDX,THETA,
     +				NP,ICHECK,MXN,MXP)
*
*       Defines nonlinear functions of site attributes. 
*	ARGUMENTS:
*	^^^^^^^^^^
*	TITLE		- Model description, for identification in
*			  programs where more than 1 model is involved.
*			  CHARACTER, input.
*       SITINF        - array of properties for each site. DOUBLE
*			  PRECISION, input/output.
*       NSITES          - no. of sites. INTEGER, input.
*       ATTRTXT         - text describing attributes at each site.
*			  CHARACTER, input/output.
*       NATTR           - number of available attributes. INTEGER,
*			  input/output.
*	CODES		- vector of covariate codes for the model.
*			  INTEGER, input.
*	XFM		- the choice of transformation. INTEGER, input
*	FOUIDX		- Indices for positions in THETA where lower
*			  limits of windows for Fourier representations
*			  are defined. FOUIDX(J) references the THETA
*			  position corresponding to attribute J. INTEGER,
*			  input.
*	LEGIDX		- Ditto, windows for Legendre representations.
*	THETA		- parameters in nonlinear transformations.
*			  DOUBLE PRECISION, input.
*	NP		- no. of parameters in the model. INTEGER, input.
*       ICHECK          - Indicator for whether parametrisation has been
*                         checked. INTEGER, input.
*	MXN		- Used in dimensioning SITINF. INTEGER, input.
*	MXP		- Used in dimensioning other arguments. INTEGER,
*			  input.
******************************************************************************
      INTEGER NP(10),ICHECK,MXN,MXP
      INTEGER NSITES,NATTR,CODES(MXP),XFM(MXP,2)
      INTEGER FOUIDX(MXP),LEGIDX(MXP)
      DOUBLE PRECISION SITINF(MXN,MXP,0:3),THETA(MXP,3)
      CHARACTER*70 ATTRTXT(MXP),TITLE
******************************************************************************
*	Additional INTEGERs
*	^^^^^^^^^^^^^^^^^^^
*	NXATTR		- number of transformed attributes
*	FREQNO		- Frequency for Fourier transform
*	PIDEF		- Indicator for whether we've calculated PI
*	DEGREE		- Degree of polynomial being calculated
*	I,J,K		- Counters
******************************************************************************
      INTEGER NXATTR,FREQNO,PIDEF,DEGREE
      INTEGER I,J,K
******************************************************************************
*	Additional DOUBLEs
*	^^^^^^^^^^^^^^^^^^
*	PI		- -i log(-1)
*	X		- value of normalized attribute for orthogonal
*			  xfrmations
*	LPK   }		- Legendre polynomial values for degree K, K-1
*	LPKM1 }		  and K-2 (for recursions)
*	LPKM2 }
*	TMP*		- Temporary storage
******************************************************************************
      DOUBLE PRECISION PI,X,LPK,LPKM1,LPKM2,TMP1,TMP2

      SAVE PI,PIDEF
      DATA PIDEF /0/
*
*	Compute PI if we haven't done so already
*
      IF (PIDEF.EQ.0) THEN
       PI = 4.0D0*DATAN(1.0D0)
       PIDEF = 1
      ENDIF
      NXATTR = NATTR
*
*	We go through the site-specific covariates one at a time, 
*	checking to see if there's a corresponding entry in XFM.
*	If so we calculate the appropriate transformation of that
*	covariate, store as an extra site attribute and re-index
*	the covariate codes so they point to the right place. Also,
*	we compute the derivatives wrt any unknown parameters. Note
*	that XFM(1,.) will always contain the number of the 
*	untransformed site attribute; CODES will point to the
*	transformed attribute. NB we always check that we've
*	defined the right number of parameters for the transformation
*	requested.
*
      DO 500 J=1,NP(1)
       IF (XFM(J,2).EQ.0) THEN
        IF (ICHECK.EQ.0) CALL THTCHK(THETA,J,0,MXP)
        GOTO 500
       ENDIF
       NXATTR = NXATTR + 1
*
*       1: Box-Cox power family (1 parameter -> 1 derivative). Have to
*	treat it separately when the parameter is zero.
*
       IF (XFM(J,2).EQ.1) THEN
        IF (ICHECK.EQ.0) CALL THTCHK(THETA,J,1,MXP)
        ATTRTXT(NXATTR) = 'Box-Cox power transform of '//
     +					ATTRTXT(XFM(J,1))
        DO 100 I=1,NSITES
         IF (SITINF(I,XFM(J,1),0).LE.0.0D0) GOTO 920
         IF (DABS(THETA(J,1)).GT.0.0D0) THEN
          SITINF(I,NXATTR,0) = 
     +    ( (SITINF(I,XFM(J,1),0)**THETA(J,1)) - 1.0D0 )/THETA(J,1)
          SITINF(I,NXATTR,1) = ( 1.0D0 + (
     +		(SITINF(I,XFM(J,1),0)**THETA(J,1))* (
     +          (THETA(J,1)*DLOG(SITINF(I,XFM(J,1),0))) - 1.0D0
     +                                  	      )
     +  		         )         ) / 
     + 	  	(THETA(J,1)**2)
         ELSE
          SITINF(I,NXATTR,0) = DLOG(SITINF(I,XFM(J,1),0))
          SITINF(I,NXATTR,1) = 0.5D0*
     +           	(DLOG(SITINF(I,XFM(J,1),0))**2)
         ENDIF
 100    CONTINUE
*
*       2: e^{ax} (1 parameter -> 1 derivative)
*
       ELSEIF (XFM(J,2).EQ.2) THEN
        IF (ICHECK.EQ.0) CALL THTCHK(THETA,J,1,MXP)
        ATTRTXT(NXATTR) = 'exp[ax] transform of '//
     +					ATTRTXT(XFM(J,1))
        DO 120 I=1,NSITES
         SITINF(I,NXATTR,0) = 
     + 		DEXP(THETA(J,1)*SITINF(I,XFM(J,1),0))
         SITINF(I,NXATTR,1) = SITINF(I,XFM(J,1),0) * 
     +                                 SITINF(I,NXATTR,0)
 120    CONTINUE
*
*       3: arctan transform (2 parameters -> 2 derivatives)
*
       ELSEIF (XFM(J,2).EQ.3) THEN
        IF (ICHECK.EQ.0) CALL THTCHK(THETA,J,2,MXP)
        ATTRTXT(NXATTR) = 'Arctan-transformed '//ATTRTXT(XFM(J,1))
        DO 140 I=1,NSITES
         SITINF(I,NXATTR,0) = 
     +		DATAN((SITINF(I,XFM(J,1),0)-THETA(J,1))/THETA(J,2))
         SITINF(I,NXATTR,1) = -THETA(J,2)/
     +	  ( (THETA(J,2)**2)+((SITINF(I,XFM(J,1),0)-THETA(J,1))**2) )
         SITINF(I,NXATTR,2) = (THETA(J,1)-SITINF(I,XFM(J,1),0))/
     +	  ( (THETA(J,2)**2)+((SITINF(I,XFM(J,1),0)-THETA(J,1))**2) )
 140    CONTINUE
*
*	11-30: Fourier series representation (parameters always
*	assumed known -> 0 derivatives). NB the idea here is for
*	orthogonal series expansions, so we only need 1 parameter
*	set (stored in THETA(FOUIDX(XFM(J,1)),*). Treat odds (sine
*	terms) and evens (cos terms) separately
*
       ELSEIF ( (XFM(J,2).GE.11).AND.(XFM(J,2).LE.30) ) THEN
        IF (FOUIDX(XFM(J,1)).EQ.0) GOTO 910
        IF (ICHECK.EQ.0) CALL THTCHK(THETA,FOUIDX(XFM(J,1)),2,MXP)
*
*	Integer arithmetic gives sequence 1,1,2,2,3,3,4,4 etc.
*
        FREQNO = (XFM(J,2)-9)/2
        TMP1 = 2.0D0*PI*DBLE(FREQNO)
        TMP2 = THETA(FOUIDX(XFM(J,1)),2) - THETA(FOUIDX(XFM(J,1)),1)
        IF ((2*FREQNO)+9.EQ.XFM(J,2) ) THEN
         WRITE(ATTRTXT(NXATTR),5) FREQNO,ATTRTXT(XFM(J,1))
         DO 160 I=1,NSITES
          X = SITINF(I,XFM(J,1),0)-THETA(FOUIDX(XFM(J,1)),1)
          X = X/TMP2
          SITINF(I,NXATTR,0) = DSIN( TMP1*X )
 160     CONTINUE
        ELSE
         WRITE(ATTRTXT(NXATTR),6) FREQNO,ATTRTXT(XFM(J,1))
         DO 161 I=1,NSITES
          X = SITINF(I,XFM(J,1),0)-THETA(FOUIDX(XFM(J,1)),1)
          X = X/TMP2
          SITINF(I,NXATTR,0) = DCOS( TMP1*X )
 161     CONTINUE
        ENDIF
*
*	31-40: Legendre polynomial representation (parameters always
*	assumed known -> 0 derivatives). Basically same idea as Fourier,
*	but interval of representation is (-1,1) instead of (0,1). 
*	Polynomials are evaluated using recursions rather than via
*	tables of coefficients.
*
       ELSEIF ( (XFM(J,2).GE.31).AND.(XFM(J,2).LE.40) ) THEN
        IF (LEGIDX(XFM(J,1)).EQ.0) GOTO 910
        IF (ICHECK.EQ.0) CALL THTCHK(THETA,LEGIDX(XFM(J,1)),2,MXP)
        TMP1 = THETA(LEGIDX(XFM(J,1)),2) + THETA(LEGIDX(XFM(J,1)),1)
        TMP2 = THETA(LEGIDX(XFM(J,1)),2) - THETA(LEGIDX(XFM(J,1)),1)
        DEGREE = XFM(J,2)-30
        WRITE(ATTRTXT(NXATTR),7) DEGREE,ATTRTXT(XFM(J,1))
        DO 180 I=1,NSITES
         X = ( (2.0D0*SITINF(I,XFM(J,1),0)) - TMP1 ) / TMP2
         IF (DEGREE.EQ.1) THEN         
          SITINF(I,NXATTR,0) = X
         ELSE
          LPKM2 = 1.0D0
          LPKM1 = X
          DO 185 K=2,DEGREE
           LPK = ( DBLE((2*K)-1)*X*LPKM1 ) - ( DBLE(K-1)*LPKM2 )
           LPK = LPK / DBLE(K)
           LPKM2 = LPKM1
           LPKM1 = LPK
 185      CONTINUE
          SITINF(I,NXATTR,0) = LPK
         ENDIF
 180    CONTINUE
       ENDIF

       CODES(J) = NXATTR

 500  CONTINUE

      RETURN
*
*	Error trapping
*
 910  WRITE(*,2) J,TITLE
      STOP
 920  WRITE(*,3) XFM(J,1),I

 2    FORMAT(/'****ERROR**** Wrong number of parameters supplied ',
     +'to ATTRXFM',/,
     +'(covariate ',I2,'). Check model definition file. ',
     +'(I''m currently working',/,
     +'on the following model:',/A70,')')
 3    FORMAT(/'****ERROR**** you have asked for a Box-Cox transform ',
     +'of site attribute ',I2,',',/,'but the value of this ',
     +'attribute is no strictly positive at site ',I3,'.') 
 5    FORMAT('Fourier sine component ',I2,' for ',A40)
 6    FORMAT('Fourier cosine component ',I2,' for ',A38)
 7    FORMAT('Legendre polynomial ',I2,' for ',A43)
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE TRENDSB(TRENDFN,YEAR,COVNO,CHOICE,THETA,ICHECK,MXP)
*
*       Returns the value of a long-term trend in year YEAR,
*	parameterised by the appropriate elements of THETA, which
*	are in row COVNO. Also check (as far as possible) that
*	parameters have been correctly specified. The 0th element
*	of TRENDFN contains the value of the function, the
*	1st and second contain values of derivatives wrt any
*	nonlinear parameters we're trying to estimate. TMP
*	is temporary storage for saving time.
*
      INTEGER YEAR,CHOICE,MXP,COVNO,ICHECK
      DOUBLE PRECISION TRENDFN(0:2),PI,THETA(MXP,3),TMP
 
      PI = 4.0D0*DATAN(1.0D0)
      IF (CHOICE.EQ.1) THEN
*
*	Linear trend - no parameters
*
       IF (ICHECK.EQ.0) CALL THTCHK(THETA,COVNO,0,MXP)
       TRENDFN(0) = (DBLE(YEAR)-1950.0D0)/1.0D1
       TRENDFN(1) = -1.0D-1
       TRENDFN(2) = 0.0D0
      ELSEIF (CHOICE.EQ.2) THEN
*
*	Piecewise linear trend (1 parameter -> 1 derivative)
*
       IF (ICHECK.EQ.0) CALL THTCHK(THETA,COVNO,1,MXP)
       IF (DBLE(YEAR).LE.THETA(COVNO,1)) THEN
	TRENDFN(0) = 0.0D0
        TRENDFN(1) = 0.0D0
       ELSE
	TRENDFN(0) = (DBLE(YEAR) - THETA(COVNO,1))/1.0D1
        TRENDFN(1) = -1.0D-1
       ENDIF
      ELSEIF (CHOICE.EQ.3) THEN
*
*	Cycle (2 parameters -> 2 derivatives)
*
       IF (ICHECK.EQ.0) CALL THTCHK(THETA,COVNO,2,MXP)
       TMP = 2.0D0*PI*(DBLE(YEAR)-THETA(COVNO,2))/THETA(COVNO,1)
       TRENDFN(0) = -DCOS(TMP)
       TRENDFN(1) = -DSIN(TMP)*TMP/THETA(COVNO,1)
       TRENDFN(2) = -2.0D0*PI*DSIN(TMP)/THETA(COVNO,1)
      ELSE
       WRITE(*,1)
       STOP
      ENDIF
      RETURN  

 1    FORMAT(/'****ERROR**** Illegal trend requested in TRENDFN')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE EXTXST(FILNO,TSCALE,PREDNO,LAG,LABEL)
*
*     To set descriptive text for external predictors. Arguments:
*
*     FILNO   - channel number of (open) file to read. Input
*     TSCALE  - Time scale (1=annual,2=minthly,3=daily). Input
*     PREDNO  - Number of predictor in the file. Input
*     LAG     - Lag of predictor value (in units of TSCALE) relative
*               to current observation.
*     LABEL   - Descriptive text. Output.
******************************************************************************
      INTEGER FILNO,TSCALE,PREDNO,LAG
      CHARACTER*70 LABEL
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     NHEAD   - Numbers of header rows to skip in each type of file
*     NPRDEF  - Number of predictors defined in the file
*     CURLIN  - current line number
*     I       - counter
******************************************************************************
      INTEGER NHEAD(3),NPRDEF,CURLIN,I
******************************************************************************
*     Additional CHARACTERs
*     ^^^^^^^^^^^^^^^^^^^^^
*     SCLTXT  - Text describing the time scale - for error message output
*     TMPSTR  - For reading/writing labels
******************************************************************************
      CHARACTER*8 SCLTXT(3)
      CHARACTER*70 TMPSTR

      DATA (NHEAD(I),I=1,3) /39,40,41/
      DATA (SCLTXT(I),I=1,3) /'annual','monthly','daily'/
*
*     Skip over the file header
*
      DO 50 I=1,NHEAD(TSCALE)
       CURLIN = I
       READ(FILNO,*,ERR=98)
 50   CONTINUE

*
*     Read no. of defined predictors, and check that the one 
*     requested is valid.
*
      CURLIN = CURLIN + 1
      READ(FILNO,*,ERR=98) NPRDEF
      IF (PREDNO.GT.NPRDEF) GOTO 99
      IF (NPRDEF.GT.950) GOTO 97
*
*     Skip to the line containing the text we want
*
      DO 60 I=1,PREDNO-1
       CURLIN = CURLIN + 1
       READ(FILNO,*,ERR=98)
 60   CONTINUE
*
*     Read the text ...
*
      CURLIN = CURLIN + 1
      READ(FILNO,'(A70)',ERR=98) LABEL
*
*     If there's a non-zero lag involved, amend text 
*
      IF (LAG.NE.0) THEN
       WRITE(TMPSTR,'(''Lag '',I3,'': '',A61)') LAG,LABEL
       READ(TMPSTR,'(A70)') LABEL
      ENDIF
*
*     ... and reposition the file (not worth thinking about
*     tracking this to improve efficiency - it only gets done a couple 
*     of times during the entire program)
*
      REWIND(FILNO)

      RETURN
*
*     Error trapping
*
 97   WRITE(*,7) SCLTXT(TSCALE)
 98   WRITE(*,5) SCLTXT(TSCALE),CURLIN
      STOP
 99   WRITE(*,6) SCLTXT(TSCALE)
      STOP

 5    FORMAT('****ERROR**** unexpected error in ',A8,
     +       ' input data file at line ',I4,'.')
 6    FORMAT('****ERROR**** you have requested an external ',
     +A8,' predictor',/,
     +'that is not defined in your input data file.')
 7    FORMAT('****ERROR**** you have defined more than 950 external ',
     +A8,' predictors.',/,'I hope you didn''t mean to do this!')
 8    FORMAT('****ERROR**** In EXTXST, scratch file is already open. ',
     +'This is a ',/,'programming error - please contact REC.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE EXTSET(FILNO,TSCALE,PREDNO,DD,MM,YY,LAG,XVAL,MISSFL)
*
*     To read the value of an `external forcing' predictor for
*     the current day/month/year, or at some time lagged relative to
*     the current one. Arguments:
*
*     FILNO   - channel number of (open) file to read. Input
*     TSCALE  - Time scale (1=annual,2=monthly,3=daily). Input
*     PREDNO  - Number of predictor in the file. Input
*     DD      - day of current month. Input
*     MM      - current month. Input
*     YY      - current year. Input.
*     LAG     - lag relative to the current day/month/year (measured
*               in units of days,months or years depending on value of
*               TSCALE).
*     XVAL    - Predictor value. Output.
*     MISSFL  - Flag to indicate mssing data. Output.
******************************************************************************
      INTEGER FILNO,TSCALE,PREDNO,DD,MM,YY,LAG,MISSFL
      DOUBLE PRECISION XVAL
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     MAXNPR  - Maximum number of external predictors
*     NHEAD   - Numbers of header rows to skip in each type of file
*     CHKFIL  - Indicators for whether input files have been checked. 
*     NPRDEF  - Number of predictors defined in the file
*     LIN1ID  - Index numbers of the first data line in each file
*     REQIDX  - Index numbers of required data line
*     BEGLIN  - Line number of first data line in each file
*     ENDLIN  - Last line number in each file
*     OLDIDX  - Index number of last row read
*     CURLIN  - Last line number read, in each file
*     REQLIN  - Required line numbers
*     FY,FM,FD- year, month and day from input file
*     IFAIL   - Error trapping flag
*     I       - counter
******************************************************************************
      INTEGER MAXNPR
      PARAMETER(MAXNPR=50)
      INTEGER NHEAD(3),CHKFIL(3),NPRDEF(3),LIN1ID(3),REQIDX(3)
      INTEGER CURLIN(3),REQLIN(3),BEGLIN(3),ENDLIN(3),OLDIDX
      INTEGER FY,FM,FD,IFAIL,I
******************************************************************************
*     Additional CHARACTERs
*     ^^^^^^^^^^^^^^^^^^^^^
*     SCLTXT  - Text describing the time scale - for error message output
******************************************************************************
      CHARACTER*8 SCLTXT(3)
******************************************************************************
*     Additional DOUBLEs
*     ^^^^^^^^^^^^^^^^^^
*     TMPARR  - To contain a row of predictors from the input file 
******************************************************************************
      DOUBLE PRECISION TMPARR(MAXNPR,3)

      DATA (NHEAD(I),I=1,3) /39,40,41/
      DATA (CHKFIL(I),I=1,3) /3*0/
      DATA (BEGLIN(I),I=1,3) /3*0/
      DATA (ENDLIN(I),I=1,3) /3*0/
      DATA (SCLTXT(I),I=1,3) /'annual','monthly','daily'/
      SAVE TMPARR,CHKFIL,LIN1ID,CURLIN,BEGLIN,ENDLIN,NPRDEF

      MISSFL = 0

*
*     Check that temporary storage is adequate
*
      IF (PREDNO.GT.MAXNPR) GOTO 96
*
*     If this is the first time we've read this file, scan through it once
*     to identify the first and last lines, and check that all the rows 
*     are present and in the right order. Other checks were performed when 
*     setting the text for the predictors.
*
      IF (CHKFIL(TSCALE).EQ.0) THEN
       CHKFIL(TSCALE) = 1
       DO 50 I=1,NHEAD(TSCALE)
        READ(FILNO,*) 
 50    CONTINUE
       CURLIN(TSCALE) = NHEAD(TSCALE) + 1
       READ(FILNO,*,ERR=98) NPRDEF(TSCALE)
       DO 60 I=1,NPRDEF(TSCALE)+1
        CURLIN(TSCALE) = CURLIN(TSCALE) + 1
        READ(FILNO,*,ERR=98)
 60    CONTINUE
*
*     Scan through the file to (a) set line numbers of first and last
*     lines (b) check that all rows are present and in the correct order
*
 70    CURLIN(TSCALE) = CURLIN(TSCALE) + 1
       CALL RDXREC(FILNO,TSCALE,FY,FM,FD,NPRDEF(TSCALE),
     +                                             TMPARR,MAXNPR,IFAIL)
       IF (IFAIL.EQ.1) GOTO 98
       IF (BEGLIN(TSCALE).EQ.0) THEN
        BEGLIN(TSCALE) = CURLIN(TSCALE)
        CALL IDXSET(TSCALE,FY,FM,FD,LIN1ID(TSCALE))
        REQIDX(TSCALE) = LIN1ID(TSCALE)
        OLDIDX = LIN1ID(TSCALE) - 1
       ENDIF
       IF (IFAIL.NE.2) THEN
        CALL IDXSET(TSCALE,FY,FM,FD,REQIDX(TSCALE))
        IF (REQIDX(TSCALE).NE.OLDIDX+1) GOTO 97
        OLDIDX = REQIDX(TSCALE)
        GOTO 70
       ENDIF
*
*     Only get here if IFAIL=2 i.e. we've reached EOF. After EOF the file
*     position is indeterminate, so rewind to be sure that pointers etc.
*     are in the right place.
*
       ENDLIN(TSCALE) = CURLIN(TSCALE) - 1
       IF (ENDLIN(TSCALE).LT.BEGLIN(TSCALE)) GOTO 99
       REWIND(FILNO)
       CURLIN(TSCALE) = 0
      ENDIF
*
*     Right: now compute the index number of the required date,
*     and hence the line number.
*
      CALL IDXSET(TSCALE,YY,MM,DD,REQIDX(TSCALE))
      REQIDX(TSCALE) = REQIDX(TSCALE) - LAG
      REQLIN(TSCALE) = BEGLIN(TSCALE)+REQIDX(TSCALE)-LIN1ID(TSCALE)

*
*     First option: it's the same line as last time. If so, copy value
*     in, and scarper.
*
 100  IF (REQLIN(TSCALE).EQ.CURLIN(TSCALE)) THEN
       XVAL = TMPARR(PREDNO,TSCALE)
       IF (DABS(XVAL+9999.9D0).LT.1.0D-4) MISSFL = 1
       RETURN       
*
*     Alternatively: maybe we already know the required line isn't in
*     the file.
*
      ELSEIF ( (REQIDX(TSCALE).LT.LIN1ID(TSCALE)).OR.
     +         (REQLIN(TSCALE).GT.ENDLIN(TSCALE)) ) THEN
       XVAL = -9999.9D0
       MISSFL = 1
       RETURN
      ENDIF
*
*     OK, now if the required line is beyond the current point in the 
*     file, read up to and including the line before it (if possible):
*
      IF (REQLIN(TSCALE).GT.CURLIN(TSCALE)) THEN
       DO 110 I = CURLIN(TSCALE)+1,REQLIN(TSCALE)-1
        READ(FILNO,*)
 110   CONTINUE
*
*     If the requested line is *earlier* than the last thing we read,
*     EITHER backspace OR rewind, depending on which is quicker (NB for
*     backspace: CURLIN holds number of last row read. E.g. if last row
*     read was 5 and we want to read row 3, we need to backspace row 5,
*     row 4 and then row 3 so that we end up reading row 3 again. 
*     CURLIN then gets set to 2).
*
 115  ELSE
       IF (2*REQLIN(TSCALE).GT.CURLIN(TSCALE)) THEN
        DO 120 I = CURLIN(TSCALE),REQLIN(TSCALE),-1
         BACKSPACE(FILNO)
 120    CONTINUE
       ELSE
        REWIND(FILNO)
        DO 125 I = 1,REQLIN(TSCALE)-1
         READ(FILNO,*)
 125    CONTINUE
       ENDIF
       CURLIN(TSCALE) = REQLIN(TSCALE)-1
      ENDIF
*
*     No need to check any more - file has already been checked on 
*     on the first pass
* 
      CURLIN(TSCALE) = REQLIN(TSCALE)
      CALL RDXREC(FILNO,TSCALE,FY,FM,FD,NPRDEF(TSCALE),
     +                                           TMPARR,MAXNPR,IFAIL)
      XVAL = TMPARR(PREDNO,TSCALE)
      IF (DABS(XVAL+9999.9D0).LT.1.0D-4) MISSFL = 1
      RETURN

*
*     Error trapping
*
 96   WRITE(*,4)
      STOP
 97   WRITE(*,5) CURLIN(TSCALE),SCLTXT(TSCALE)
      STOP
 98   WRITE(*,6) SCLTXT(TSCALE),CURLIN(TSCALE)
      STOP
 99   WRITE(*,7) SCLTXT(TSCALE)
      STOP

 4    FORMAT('****ERROR**** Inadequate storage in routine EXTSET ',
     +'(file funcdefs.f).',/,'Please increase value of MAXNPR ',
     +'and recompile.')
 5    FORMAT('****ERROR**** line ',I5,' of ',A8,' input data',
     +' file carries incorrect date.')
 6    FORMAT('****ERROR**** unexpected error in ',A8,
     +       ' input data file at line ',I5,'.')
 7    FORMAT('****ERROR**** no data in ',A8,' input data file.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE RDXREC(FILNO,TSCALE,FY,FM,FD,NPRDEF,XVALS,MXNP,IFAIL)
*
*     Reads a record of an `external' data file. Arguments:
*
*     FILNO    - Channel number of open file. Input
*     TSCALE   - 1 for annual, 2 for monthly and 3 for daily data. Input
*     FY       - year read from file. Output
*     FM       - month read from file. Output
*     FD       - day read from file. Output
*     NPRDEF   - Number of predictors defined. Input
*     XVALS    - array of predictors read from file. Output
*     MXNP     - Maximum number of predictors. Input.
*     IFAIL    - error flag. Output
******************************************************************************
      INTEGER FILNO,TSCALE,FY,FM,FD,NPRDEF,MXNP,IFAIL
      DOUBLE PRECISION XVALS(MXNP,3)
      INTEGER I

      FY = 0
      FM = 0
      FD = 0
      IFAIL = 0
      IF (TSCALE.EQ.1) THEN
       READ(FILNO,1,ERR=99,END=98,IOSTAT=IFAIL) 
     +                         FY,(XVALS(I,TSCALE),I=1,NPRDEF)
      ELSEIF (TSCALE.EQ.2) THEN
       READ(FILNO,2,ERR=99,END=98,IOSTAT=IFAIL) 
     +                         FY,FM,(XVALS(I,TSCALE),I=1,NPRDEF)
      ELSEIF (TSCALE.EQ.3) THEN
       READ(FILNO,3,ERR=99,END=98,IOSTAT=IFAIL) 
     +                         FY,FM,FD,(XVALS(I,TSCALE),I=1,NPRDEF)
      ENDIF
      RETURN
*
*     Error and EOF trapping
*
 98   IFAIL = 2
      RETURN
 99   IFAIL = 1
      RETURN

 1    FORMAT(I4,T10,50(F10.0))
 2    FORMAT(I4,1X,I2,T10,50(F10.0))
 3    FORMAT(I4,2I2,1X,50(F10.0))
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE IDXSET(TSCALE,YY,MM,DD,IDX)
*
*     Allocates unique indices to individual years, months or days.
*     Arguments:
*     TSCALE     1 for annual, 2 for monthly and 3 for daily data. Input
*     YY         Year. Input
*     MM         Month. Input
*     DD         Day. Input
*     IDX        Index number (it's the number of time units since
*                1st January 2000 i.e. 1/1/2000 gets an index of zero
*                if TSCALE = 3, January 2000 gets an index of zero if
*                TSCALE = 2, and 2000 gets an index of zero if
*                TSCALE = 3). Output
******************************************************************************
      INTEGER TSCALE,YY,MM,DD,IDX
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     JULBL      Julian day number of baseline date (daily data only)
*     JULREQ     Julian day number of required date.
*     IFAIL      Return exit code for Julian conversion routines
******************************************************************************
      INTEGER JULBL,JULREQ,IFAIL

*
*     1st Jan 2000 was Julian day 2451545
*
      DATA JULBL /2451545/
      SAVE JULBL

      IF (TSCALE.EQ.1) THEN
       IDX = YY-2000
      ELSEIF (TSCALE.EQ.2) THEN
       IDX = (12*(YY-2000)) + MM - 1
      ELSEIF (TSCALE.EQ.3) THEN
       CALL DMYTOJ(DD,MM,YY,JULREQ,IFAIL)
       IDX = JULREQ - JULBL
      ENDIF

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE MONSB(MONFN,MONTH,YEAR,CHOICE)
*
*       On exit, the variable MONFN contains the value of a 
*	monthly effect for month MONTH, year YEAR
*
      DOUBLE PRECISION MONFN,PI
      INTEGER MONTH,YEAR,CHOICE,PICALC

      DATA PICALC /0/
      SAVE PICALC,PI

      IF (PICALC.EQ.0) THEN
       PI = 4.0D0*DATAN(1.0D0)
       PICALC = 1
      ENDIF
*
*	Sine waves
*
      IF (CHOICE.EQ.1) THEN
       MONFN=DCOS(DBLE(MONTH)*PI/6.0D0)
      ELSEIF (CHOICE.EQ.2) THEN
       MONFN=DSIN(DBLE(MONTH)*PI/6.0D0)
      ELSEIF (CHOICE.EQ.3) THEN
       MONFN=DCOS(DBLE(MONTH)*PI/3.0D0)
      ELSEIF (CHOICE.EQ.4) THEN
       MONFN=DSIN(DBLE(MONTH)*PI/3.0D0)
*
*	Indicators for individual months
*
      ELSEIF ( (CHOICE.GE.11).AND.(CHOICE.LE.22) ) THEN
       IF (MONTH.EQ.CHOICE-10) THEN
        MONFN = 1.0D0
       ELSE
        MONFN = 0.0D0
       ENDIF
      ELSE
       WRITE(*,1)
       STOP
      ENDIF
 1    FORMAT('****ERROR**** Illegal choice of predictor in MONFN')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE PRVTXT(COVCODE,XLABEL)
*
*     Produces a text label for a predictor which is a nonlinear
*     transformation of some previous day's values. Arguments:
*     
*     COVCODE   - coding for the transformation, in the form (1000x)+d,
*                 where x is the transformation to be used, and d is 
*                 the lag
*     XLABEL    - The text label. Input/Output
******************************************************************************
      INTEGER COVCODE
      CHARACTER XLABEL*70
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     LAG       - number of days ago
*     XFRM      - index of the transformation
*     WTSCHM    - Weighting scheme, for use when a weighted average
*                 of previous days' values at different sites is
*                 being used
*     WXORDR    - Order in which averaging and transformation are
*                 performed
*     TXTLEN    - length of text label
*     XFMLEN    - length of transformation text (see XFMTXT below)
*     WTXLEN    - length of weighted average text (see WTXT below)
*     I,J       - counters
******************************************************************************
      INTEGER LAG,XFRM,WTSCHM,WXORDR
      INTEGER TXTLEN,XFMLEN(0:10,2),WTXLEN(0:10,2)
      INTEGER I,J
******************************************************************************
*     Additional CHARACTERs
*     ^^^^^^^^^^^^^^^^^^^^^
*     XFMTXT    - Text for transformations - first element is to go
*                 at the beginning, last at the end
*     WTXT      - Text for weighted averages of previous days' values
*     TMPTX?    - Temporary storage
******************************************************************************
      CHARACTER*40 XFMTXT(0:10,2),WTXT(0:10,2)
      CHARACTER*40 TMPTX1,TMPTX2,TMPTX3,TMPTX4

      DATA ((XFMTXT(I,J),J=1,2),I=0,5) /'','','Ln(',')',
     +'Ln(1+',')','I(','>0)','Trace indicator: ','',
     +'I(Y[t-k]>0: k=1 to ',')'/
      DATA ((XFMLEN(I,J),J=1,2),I=0,5) /2*0,3,1,5,1,2,3,17,0,19,1/
      DATA ((WTXT(I,J),J=1,2),I=0,3) /'','','Mean of ','',
     +'Distance-weighted mean of ','',
     +'Shift/distance-weighted mean of ',''/
      DATA ((WTXLEN(I,J),J=1,2),I=0,3) /2*0,8,0,26,0,32,0/
*
*     Interpret COVCODE (integer arithmetic)
*
      LAG = MOD(COVCODE,1000)
      XFRM = MOD(COVCODE/1000,10)
      WTSCHM = MOD(COVCODE/10000,10)
      WXORDR = COVCODE/100000
*
*     Find out how long the text is at present
*
      TXTLEN = INDEX(XLABEL,'  ') - 1
      IF (TXTLEN.EQ.-1) TXTLEN = LEN(XLABEL)
*
*     Now set text (default core label needs changing for
*     persistence indicators). First the case when we're
*     doing averages of transformations ...
*
      IF (WXORDR.EQ.0) THEN
       IF (XFRM.EQ.5) THEN
        WRITE(XLABEL,'(I2)') LAG 
        TXTLEN = INDEX(XLABEL,'  ') - 1
       ENDIF

       IF (XFRM.LE.5) THEN
        TMPTX1 = WTXT(WTSCHM,1)
        TMPTX2 = XFMTXT(XFRM,1)
        TMPTX3 = XFMTXT(XFRM,2)
        TMPTX4 = WTXT(WTSCHM,2)
        XLABEL = TMPTX1(1:WTXLEN(WTSCHM,1)) // 
     +       TMPTX2(1:XFMLEN(XFRM,1)) // 
     +       XLABEL(1:TXTLEN) // 
     +       TMPTX3(1:XFMLEN(XFRM,2)) //
     +       TMPTX4(1:WTXLEN(WTSCHM,2))
       ELSE
        WRITE(*,1)
        STOP
       ENDIF
*
*     And now the case where we're doing transformations of averages ...
*
      ELSEIF (WXORDR.EQ.1) THEN
       IF (XFRM.LE.5) THEN
        TMPTX1 = XFMTXT(XFRM,1)
        TMPTX2 = WTXT(WTSCHM,1)
        TMPTX3 = WTXT(WTSCHM,2)
        TMPTX4 = XFMTXT(XFRM,2)
        IF (XFRM.LE.4) THEN
         XLABEL = TMPTX1(1:XFMLEN(XFRM,1)) // 
     +       TMPTX2(1:WTXLEN(WTSCHM,1)) // 
     +       XLABEL(1:TXTLEN) // 
     +       TMPTX3(1:WTXLEN(WTSCHM,2)) //
     +       TMPTX4(1:XFMLEN(XFRM,2))
*
*     `Persistence' requires special treatment to create a sensible
*     label
*
        ELSE
         WRITE(XLABEL,'(I2)') LAG 
         TXTLEN = INDEX(XLABEL,'  ') - 1
         XLABEL = 'I(' // TMPTX2(1:WTXLEN(WTSCHM,1)) //
     +          'Y[t-k] > 0: k=1 to ' // XLABEL(1:TXTLEN) // ')'
        ENDIF
       ELSE
        WRITE(*,1)
        STOP
       ENDIF
      ELSE
       WRITE(*,1)
       STOP
      ENDIF

 1    FORMAT('****ERROR**** Illegal transformation found in PRVTXT')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE DYSET(DD,MM,YY,PREV,SITE,NSITES,TRACE,COVCODE,
     +                  THETA,PWTIDX,SITINF,ICHECK,DYPRED,MXN,MXP)
******************************************************************************
*       Returns a predictor that varies on a daily timescale. Arguments
*       (all input):
*
*       DD       - Day of month. 
*       MM       - Month of year. 
*       YY       - Year. 
*       SITE     - Site we're working on.
*       NSITES   - Total number of sites
*       PREV     - Vector of up to 10 previous values at each site. 
*       TRACE    - Trace threshold.
*       COVCODE  - AS for PRVTXT above
*       THETA    - Parameters in nonlinear transofrmations (for computing
*                  distance-dependent weights)
*       PWTIDX   - Index locating row of THETA to use for calculating
*                  distance-dependent weights.
*       SITINF   - Raw site information (for calculating inter-site 
*                  distances)
*       ICHECK   - Indicator to tell us whether the model parametrisation
*                  has been checked
*       DYPRED   - OUTPUT: value of predictor, together with any
*                  derivatives wrt nonlinear parameters.
*       MXN      - Maximum number of sites.
*       MXP      - Maximum number of site attributes
******************************************************************************
      INTEGER DD,MM,YY,SITE,NSITES,COVCODE,MXN,MXP
      INTEGER ICHECK,PWTIDX(MXP,0:3)
      DOUBLE PRECISION PREV(MXN,10),TRACE,DYPRED(0:3)
      DOUBLE PRECISION THETA(MXP,3),SITINF(MXN,MXP,0:3)
******************************************************************************
*     Additional PARAMETER
*     ^^^^^^^^^^^^^^^^^^^^
*     MXN2       - For dimensioning WGTARR array. Should be set equal
*                  to value of MXN in main program
******************************************************************************
      INTEGER MXN2
      PARAMETER(MXN2=200)
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     CODE       - code used to define predictor
*     XFRM       - transformation, if any
*     WTSCHM     - Weighting scheme for averages of previous days' values
*     WXORDR     - Flag indicating order of averaging and transformation
*     NPREQ      - Number of parameters required by various weighting schemes
*     NDERIV     - Number of derivatives we want to calculate
*     IDERIV     - Indicator - do we want to calculate derivatives of
*                  transformations?
*     PICALC     - have we calculated PI yet?
*     CUMDAY     - cumulative no. of days to beginning of 
*                  current month (calculate as though it's a leap
*                  year - in non-leap years this will result in a 
*                  tiny discontinuity in sin and cos terms, but the
*                  effect will be miniscule and it saves 
*                  on computation time)
*     NDAYS      - numbers of days in each month (ditto)
*     CURDAY     - number of current day (1-366)
*     MONADJ     - Month for which an adjustment is to be made 
*     I,J        - counters
******************************************************************************
      INTEGER I,J,CODE,XFRM,PICALC,WTSCHM,WXORDR,NPREQ(10)
      INTEGER NDERIV,IDERIV,CUMDAY(12),NDAYS(12),CURDAY,MONADJ
******************************************************************************
*     Additional DOUBLEs
*     ^^^^^^^^^^^^^^^^^^
*     PI         - the usual
*     TRIGX      - argument to a trigonometric function
*     TMP        - temporary storage
*     PREVEC     - Vector (as opposed to matrix) of previous values
*     SUMWT      - Sum of weights in spatial averages
*     CURWT      - Current weight, together with any necessary 
*                  derivatives wrt unknown parameters
*     SUMWT      - Sum of weights and their derivatives
*     WGTARR     - Array used for derivative calculation (the weights 
*                  themselves are quotients - we need to find the sums of
*                  the raw weights and their derivatives before we can
*                  differentiate the normalised weights)
******************************************************************************
      DOUBLE PRECISION PI,TRIGX,TMP(0:1),PREVEC(10)
      DOUBLE PRECISION CURWT(0:3),SUMWT(0:3),WGTARR(MXN2,0:3)

      DATA PICALC /0/
      DATA (CUMDAY(I),I=1,12) /0,31,60,91,121,152,182,
     +                             213,244,274,305,335/
      DATA (NDAYS(I),I=1,12) /31,29,31,30,31,30,31,31,30,31,30,31/
      DATA (NPREQ(I),I=1,3) /0,1,3/
      SAVE PICALC,PI

      IF (PICALC.EQ.0) THEN
       PI = 4.0D0*DATAN(1.0D0)
       PICALC = 1
      ENDIF
*
*     Interpret COVCODE
*
      CODE = MOD(COVCODE,1000)
*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
*     Daily effects representing autocorrelation:
*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF (CODE.LE.10) THEN
       XFRM = MOD(COVCODE/1000,10)
       WTSCHM = MOD(COVCODE/10000,10)
       WXORDR = COVCODE/100000
       IDERIV = 0
*
*       Missing values - set to large negative number so we can catch
*       them. NB this is allowed for in fitting progs anyway, where 
*	there are flags. Any case is defined as valid if it has the
*       required number of previous days' values at the site of interest
*       (so, for predictors that are spatial averages, there could
*       be many missing values in the final `average' and the program 
*       will still accept them). 
*
       IF (PREV(SITE,CODE).LT.0.0D0) THEN
	DYPRED(0) = -1.0D12
       ELSE
        IF (XFRM.EQ.5) THEN
         DO 50 I=1,CODE
          IF (PREV(SITE,I).LT.0.0D0) THEN
           DYPRED(0) = -1.0D12
           RETURN
          ENDIF
 50      CONTINUE
        ENDIF
*
*     For non-missing values, calculate transformations. First the
*     case when we're simply looking at each site's history 
*     individually:
*
        IF (WTSCHM.EQ.0) THEN
         IF (XFRM.EQ.5) THEN
          DO 100 I=1,CODE
           PREVEC(I) = PREV(SITE,I)
 100      CONTINUE        
         ENDIF
         CALL PRVXFM(PREV(SITE,CODE),XFRM,TRACE,CODE,PREVEC,
     +                                                  IDERIV,TMP)
         DYPRED(0) = TMP(0)
*
*     Now the case where we're averaging. First calculate raw weights
*     (and, if necessary, their partial derivatives). Derivatives
*     (all or none) are only computed if absolutely necessary, since they're
*     expensive. To see if we need them, look at elements of THETA in
*     `nonlinear' section.
*
        ELSE
         NDERIV = 0
         DO 105 I=0,NPREQ(WTSCHM)
          SUMWT(I) = 0.0D0
          DYPRED(I) = 0.0D0
          IF ((I.GT.0).AND.(PWTIDX(WTSCHM,I).NE.0)) 
     +                                 NDERIV = NPREQ(WTSCHM)
 105     CONTINUE
         DO 110 J = 1,NSITES
          IF (PREV(J,CODE).GE.0.0D0) THEN
           CALL SETWGT(SITE,J,WTSCHM,NPREQ,NDERIV,SITINF,
     +                       THETA,PWTIDX,CODE,ICHECK,CURWT,MXN,MXP)
           DO 109 I=0,NDERIV
            SUMWT(I) = SUMWT(I) + CURWT(I)
            WGTARR(J,I) = CURWT(I)
 109       CONTINUE
          ENDIF
 110     CONTINUE
*
*     ... then averages. Compute derivatives of normalised weights along 
*     the way. NB need to check that we're not dividing by zero. This
*     will occur if, for example, there's only 1 site. In this case it 
*     gets a weight of 1, and the correct derivative is clearly zero. 
*
         DO 120 J = 1,NSITES
          IF (PREV(J,CODE).GE.0.0D0) THEN
           CURWT(0) = WGTARR(J,0)
           DO 112 I=1,NDERIV
            IF (DABS(SUMWT(0)).GT.1.0D-12) THEN
             CURWT(I) = ( (WGTARR(J,I)*SUMWT(0)) - 
     +                    (WGTARR(J,0)*SUMWT(I)) ) / (SUMWT(0)**2)
            ELSE
             CURWT(I) = 0.0D0
            ENDIF
 112       CONTINUE
*     
*     Here are means of transformed values, and their derivatives
*
           IF (WXORDR.EQ.0) THEN
            IF (XFRM.EQ.5) THEN
             DO 114 I=1,CODE
              PREVEC(I) = PREV(J,I)
 114         CONTINUE
            ENDIF
            CALL PRVXFM(PREV(J,CODE),XFRM,TRACE,CODE,PREVEC,
     +                                                IDERIV,TMP)
            DO 116 I=0,NDERIV
             DYPRED(I) = DYPRED(I) + (CURWT(I) * TMP(0))
 116        CONTINUE
*
*     And here are transformed values of means. `Persistence' has
*     to be done separately - basically, as soon as we find a zero
*     with non-zero weight, we're out.
*
           ELSEIF (WXORDR.EQ.1) THEN 
            IF (XFRM.NE.5) THEN
             DO 117 I=0,NDERIV
              DYPRED(I) = DYPRED(I) + (CURWT(I) * PREV(J,CODE))
 117         CONTINUE
            ELSE
             DO 118 I=1,CODE
              IF ((PREV(J,I).LE.0.0D0).AND.(CURWT(0).GT.0.0D0)) THEN
               DYPRED(0) = 0.0D0
               RETURN
              ENDIF
 118         CONTINUE
            ENDIF
           ELSE
            WRITE(*,1)
            STOP
           ENDIF
          ENDIF
 120     CONTINUE
*
*     Right, now finish off. Average of transformed value just has
*     to be turned into an average ...
*
 119     IF (WXORDR.EQ.0) THEN
          DYPRED(0) = DYPRED(0)/SUMWT(0)
*
*     Transformed value of average has to be calculated, via
*     a call to PRVXFM unless it's a persistence thing (in
*     which case the only way we could have got here is if it's
*      a 1). Also, if we're estimating parameters (which will
*     happen if NERIV > 0), we'll need derivatives.
*
         ELSE
          DYPRED(0) = DYPRED(0)/SUMWT(0)
          IF (NDERIV.GT.0) IDERIV = 1
          IF (XFRM.NE.5) THEN
           CALL PRVXFM(DYPRED(0),XFRM,TRACE,CODE,PREVEC,IDERIV,TMP)
           DYPRED(0) = TMP(0)
           DO 130 I=1,NDERIV
            DYPRED(I) = TMP(1) * DYPRED(I)
 130       CONTINUE
          ELSE
           DYPRED(0) = 1.0D0
          ENDIF
         ENDIF
        ENDIF
       ENDIF
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*     Other effects
*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
*
*     Daily representation of seasonality (cos/sin of 2 * pi * w * t/366,
*     where w is 1 or 2)
*
      ELSEIF ( (CODE.GE.21).AND.(CODE.LE.24) ) THEN
       CURDAY = CUMDAY(MM) + DD       
       IF (CODE.LE.22) THEN
        TRIGX = (PI*DBLE(CURDAY)/183.0D0)
       ELSE
        TRIGX = (PI*DBLE(CURDAY)/91.5D0)
       ENDIF
       IF ((CODE.EQ.21).OR.(CODE.EQ.23)) THEN
         DYPRED(0) = DCOS(TRIGX)
       ELSE
         DYPRED(0) = DSIN(TRIGX)
       ENDIF
*
*     Smooth seasonal adjustment based on a bisquare function -
*     takes value zero on the 0th day of the required month, and on 
*     the (N+1)th day, where N is the number of days in the month
*
      ELSEIF ( (CODE.GE.31).AND.(CODE.LE.42) ) THEN
       MONADJ = CODE-30
       IF (MM.NE.MONADJ) THEN
        DYPRED(0) = 0.0D0
       ELSE
        TMP(0) = DBLE(NDAYS(MM)+1) / 2.0D0
        TMP(0) = (DBLE(DD) - TMP(0)) / TMP(0)
        DYPRED(0) = (1.0D0-(TMP(0)**2))**2
       ENDIF
      ELSE
       WRITE(*,3)
       STOP
      ENDIF 

 1    FORMAT('****ERROR**** Invalid value of WXORDR in DYSET. This ',
     +'is a ',/,'programming error - contact REC.')
 3    FORMAT('****ERROR**** Undefined effect found in DYSET') 
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE PRVXFM(X,XFRM,TRACE,LAG,PREVEC,IDERIV,FX)
*
*     Calculates transformations of previous days' values. Arguments:
*     X (double precision, input): value to be transformed
*     XFRM (integer, input):       transform to be applied
*     TRACE (double, input):       trace threshold
*     LAG (integer,input):         Lag for persistence indicators
*     PREVEC (double, input):      Vector of previous values
*     IDERIV (integer, input):     Calculate derivatives?
*     FX:                          Transformed value (output), together
*                                  with first derivative if there is one
******************************************************************************
      DOUBLE PRECISION X,TRACE,PREVEC(10),FX(0:1)
      INTEGER XFRM,LAG,IDERIV
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     I          Counter
******************************************************************************
      INTEGER I

      IF (XFRM.EQ.0) THEN
       FX(0) = X
       IF (IDERIV.EQ.1) FX(1) = 1
      ELSEIF (XFRM.EQ.1) THEN
       IF (X.GT.0.0D0) THEN
        FX(0) = DLOG(X)
        IF (IDERIV.EQ.1) FX(1) = 1.0D0/X
       ELSE
        WRITE(*,1)
        FX(0) = -1.0D12
        IF (IDERIV.EQ.1) FX(1) = -1.0D12
       ENDIF
      ELSEIF (XFRM.EQ.2) THEN
       IF (X.GT.-1.0D0) THEN
        FX(0) = DLOG(1.0D0+X)
        IF (IDERIV.EQ.1) FX(1) = 1.0D0/(1.0D0+X)
       ELSE
        WRITE(*,1)
        FX(0) = -1.0D12
        IF (IDERIV.EQ.1) FX(1) = -1.0D12
       ENDIF
      ELSEIF (XFRM.EQ.3) THEN
       IF (X.GT.0.0D0) THEN
        FX(0) = 1.0D0
       ELSE
        FX(0) = 0.0D0
       ENDIF
       IF (IDERIV.EQ.1) FX(1) = 0.0D0
      ELSEIF (XFRM.EQ.4) THEN
       IF ((X.GT.0.0D0).AND.
     +     (X.LT.TRACE)) THEN
        FX(0) = 1.0D0
       ELSE
        FX(0) = 0.0D0
       ENDIF
       IF (IDERIV.EQ.1) FX(1) = 0.0D0
      ELSEIF (XFRM.EQ.5) THEN
       FX(0) = 1.0D0
       IF (IDERIV.EQ.1) FX(1) = 0.0D0
       DO 50 I=1,LAG
        IF (PREVEC(I).LE.0.0D0) THEN
         FX(0) = 0.0D0
         RETURN
        ENDIF
 50    CONTINUE
      ELSE
       WRITE(*,2)
       STOP
      ENDIF
 1    FORMAT('****WARNING**** Attempt to log a non-positive ',
     +'number while',/,'calculating predictors from previous days')
 2    FORMAT('****ERROR**** Illegal transformation found in PRVXFM')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE SETWGT(SITE1,SITE2,WTSCHM,NPREQ,NDERIV,SITINF,
     +                        THETA,PWTIDX,LAG,ICHECK,WEIGHT,MXN,MXP)
******************************************************************************
*     Calculates weights associated with neighbours of a site, when
*     averaging previous days' values. Arguments:
*     SITE1      - Number of site at which we're predicting. INPUT
*     SITE2      - Number of neighbouring site. INPUT
*     WTSCHM     - Weighting scheme to be used. INPUT
*     NPREQ      - Number of parameters involved in various weighting 
*                  schemes. INPUT.
*     NDERIV     - Number of derivatives we'll actually calculate
*     SITINF     - Array of site information (distances will be 
*                  calculated from first 2 columns). INPUT
*     THETA      - Parameters used in nonlinear transformations. INPUT
*     PWTIDX     - Index pointing to the row of THETA that holds
*                  the parameters used in WTSCHM. INPUT
*     LAG        - Number of days back we're going. Input
*     ICHECK     - Indicator for whether the model parametrisation
*                  has yet been checked. INPUT
*     WEIGHT     - Allocated weight, together with necessary derivatives
*                  with respect to nonlinear parameters.
*     MXN        - Maximum number of sites      } Used for dimensioning.
*     MXP        - Maximum number of parameters } INPUT
******************************************************************************
      INTEGER SITE1,SITE2,WTSCHM,NPREQ(10),NDERIV,MXN,MXP
      INTEGER PWTIDX(MXP,0:3),ICHECK,LAG
      DOUBLE PRECISION SITINF(MXN,MXP,0:3),THETA(MXP,3),WEIGHT(0:3)
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     ROW        - Required row of THETA
*     I,J        - counters
******************************************************************************
      INTEGER ROW
******************************************************************************
*     Additional DOUBLEs
*     ^^^^^^^^^^^^^^^^^^
*     DISTNC     - Distance between sites
*     XSEP,YSEP  - Site separtaion in X & Y directions
******************************************************************************
      DOUBLE PRECISION DISTNC,XSEP,YSEP

*
*     First up: check that if there ought to be some parameters 
*     associated with this weighting scheme, they've been defined
*
      IF (WTSCHM.NE.1) THEN
       ROW = PWTIDX(WTSCHM,0)
       IF (ROW.EQ.0) GOTO 920
      ENDIF
*
*     Uniform weights:
*
      IF (WTSCHM.EQ.1) THEN
       WEIGHT(0) = 1.0D0
*
*     Exponentially weighted by distance - check that the required
*     parameter has been defined (and that no other parameter
*     has) and if so, compute the weight and its derivative wrt
*     the exponential decay rate.
*  
      ELSEIF (WTSCHM.EQ.2) THEN
       IF (ICHECK.EQ.0) CALL THTCHK(THETA,ROW,NPREQ(WTSCHM),MXP)
       DISTNC = DSQRT( (SITINF(SITE2,1,0)-SITINF(SITE1,1,0))**2 +
     +                 (SITINF(SITE2,2,0)-SITINF(SITE1,2,0))**2 )
       WEIGHT(0) = DEXP(-THETA(ROW,1)*DISTNC)
       IF (NDERIV.GT.0) THEN
        WEIGHT(1) = -(DISTNC*WEIGHT(0))
       ENDIF
      ELSEIF (WTSCHM.EQ.3) THEN
*
*     Shift and exponential decay. 3 parameters -> 3 derivatives.
*     Derivatives require special treatment at the origin :-(
*
       IF (ICHECK.EQ.0) CALL THTCHK(THETA,ROW,NPREQ(WTSCHM),MXP)
       XSEP = SITINF(SITE2,1,0) - SITINF(SITE1,1,0) -
     +                             (DBLE(LAG)*THETA(ROW,1))
       YSEP = SITINF(SITE2,2,0) - SITINF(SITE1,2,0) -
     +                             (DBLE(LAG)*THETA(ROW,2))
       DISTNC = DSQRT( (XSEP**2) + (YSEP**2) )
       WEIGHT(0) = DEXP(-THETA(ROW,3)*DISTNC)
       IF (NDERIV.GT.0) THEN
        IF (DISTNC.GT.1.0D-12) THEN
         WEIGHT(1) = THETA(ROW,3) * DBLE(LAG) * XSEP *
     +                                 WEIGHT(0) / DISTNC
         WEIGHT(2) = THETA(ROW,3) * DBLE(LAG) * YSEP *  
     +                                 WEIGHT(0) / DISTNC
         WEIGHT(3) = -(DISTNC*WEIGHT(0))
        ELSE
         WEIGHT(1) = -( THETA(ROW,3) * DBLE(LAG) * WEIGHT(0) )
         WEIGHT(2) = WEIGHT(1)
         WEIGHT(3) = 0.0D0
        ENDIF
       ENDIF
      ELSE
       WRITE(*,1)
       STOP
      ENDIF

      RETURN
*
*     Error trapping beyond here
*

 920  WRITE(*,3)
      STOP

 1    FORMAT(/'****ERROR**** Illegal weighting scheme requested ',
     +'in SETWGT.')
 2    FORMAT(/'****ERROR**** Wrong number of parameters supplied ',
     +'to SETWGT',/,
     +'(covariate ',I2,'). Check model definition file.')
 3    FORMAT(/'****ERROR**** You''ve asked me to compute weighted',
     +' averages of',/,'previous days'' values, but you haven''t',
     +' given me any parameters',/,'for the weighting scheme!')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE THTCHK(THETA,ROW,NPAR,MXP)
******************************************************************************
*     Checks that the correct number of nonlinear parameters have 
*     been defined. Arguments (all input):
*     THETA     - array of nonlinear parameters
*     ROW       - Row of THETA that we're looking at
*     NPAR      - Number of parameters that should have been defined
*     MXP       - Dimensioning for THETA
******************************************************************************
      INTEGER ROW,NPAR,MXP
      DOUBLE PRECISION THETA(MXP,3)
      INTEGER I

*
*     First check: all necessary parameters *have* been defined
*

      DO 10 I=1,NPAR
       IF (THETA(ROW,I).GT.1.0D8) GOTO 910
 10   CONTINUE

*
*     Second: check that no extraneous parameters have been defined
*
      DO 20 I=NPAR+1,3
       IF (THETA(ROW,I).LE.1.0D8) GOTO 910
 20   CONTINUE

      RETURN
*
*     Error trapping
*

 910  WRITE(*,1) ROW
      STOP

 1    FORMAT(/'****ERROR**** Wrong number of parameters supplied ',
     +'for transformation ',/,'of covariate ',I2,'. Check model ',
     +'definition file.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE RHOCHK(RHO,SPMOD,FILNM,MXP)
******************************************************************************
*     Checks that the correct number of parameters have been 
*     defined for a spatial dependence structure, and that they
*     satisfy any necessary constraints. Arguments (all input):
*     RHO       array of parameters
*     SPMOD     Selected dependence model
*     FILNM     Name of file from which model definition has been read
*     MXP       Dimensioning for RHO
******************************************************************************
      INTEGER SPMOD,MXP
      DOUBLE PRECISION RHO(MXP)
      CHARACTER*12 FILNM
******************************************************************************
*     Extra INTEGERs
*     --------------
*
*       NPAR    Number of parameters required for spatial dependence model
*       I       Counter
******************************************************************************
      INTEGER I,NPAR(0:40)
******************************************************************************
*     Extra DOUBLEs
*     -------------
*
*       LL,UL   Lower and upper bounds on parameters in spatial dependence
*               model
******************************************************************************
      DOUBLE PRECISION LL(40),UL(40)
******************************************************************************
      DATA (NPAR(I),I=0,40) /41*-1/
      DATA (LL(I),I=1,40) /40*-1.0D6/
      DATA (UL(I),I=1,40) /40*1.0D6/      
      
      NPAR(0) = 0
      NPAR(1) = 0
      NPAR(2) = 1
      NPAR(3) = 1
      NPAR(4) = 2
      NPAR(5) = 2
      NPAR(6) = 3
      NPAR(21) = 1
      NPAR(22) = 1

      IF (SPMOD.EQ.0) THEN
*       Do nothing!
      ELSEIF (SPMOD.EQ.1) THEN
*       Do nothing!
      ELSEIF (SPMOD.EQ.2) THEN
       LL(1) = -1.0D0
       UL(1) = 1.0D0
      ELSE IF ( (SPMOD.EQ.3).OR.(SPMOD.EQ.4) ) THEN
       LL(1) = 0.0D0
       IF (SPMOD.EQ.4) THEN
        LL(2) = -1.0D0
        UL(2) = 1.0D0
       ENDIF
      ELSE IF ( (SPMOD.EQ.5).OR.(SPMOD.EQ.6) ) THEN
       LL(1) = 0.0D0
       LL(2) = 0.0D0
       IF (SPMOD.EQ.6) THEN
        LL(3) = -1.0D0
        UL(3) = 1.0D0
       ENDIF
      ELSEIF (SPMOD.EQ.21) THEN
*       Do nothing!
      ELSEIF (SPMOD.EQ.22) THEN
*       Do nothing!
      ELSE
       WRITE(*,4) FILNM
       STOP
      ENDIF
      
      IF (NPAR(SPMOD).LT.0) THEN
       WRITE(*,1) FILNM
       STOP
      ENDIF
*
*     First check: all necessary parameters *have* been defined
*
      DO 10 I=1,NPAR(SPMOD)
       IF (RHO(I).GT.1.0D8) GOTO 910
 10   CONTINUE

*
*     Second: check that no extraneous parameters have been defined
*
      DO 20 I=NPAR(SPMOD)+1,MXP
       IF (RHO(I).LE.1.0D8) GOTO 910
 20   CONTINUE
 
*
*     Third: check that all parameters satisfy constraints
*
      DO 30 I=1,NPAR(SPMOD)
       IF ( (RHO(I).LT.LL(I)) .OR. (RHO(I).GT.UL(I)) ) GOTO 920
 30   CONTINUE

      RETURN
*
*     Error trapping
*

 910  WRITE(*,2) FILNM
      STOP
 920  WRITE(*,3) FILNM
      STOP

 1    FORMAT(/'****ERROR**** in routine RHOCHK: unknown spatial ',
     +        'dependence structure',/'in file ',A12,'.')
 2    FORMAT(/'****ERROR**** in file ',A12,', wrong number of ',
     +        'parameters',/'supplied for spatial dependence ',
     +        'structure.')
 3    FORMAT(/'****ERROR**** in file ',A12,', one or more of the ',
     +        'spatial',/'dependence structure parameters ',
     +        'is out of bounds.')
 4    FORMAT(/'****ERROR**** unknown spatial dependence structure ',
     +        'requested',/'in file ',A12,'.')
      END
